# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2019-11-25 15:34:37 UTC
import datetime

version = "3.2.2"
githash = "ce42e6e238c200a4715785ef8c9d233f612d0c75"


major = 3
minor = 2
bugfix = 2

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2019, 11, 25, 15, 34, 37)
debug = False

astropy_helpers_version = ""
