# Autogenerated by Astropy-affiliated package astropy_healpix's setup.py on 2019-11-25 15:34:39 UTC
import datetime

version = "0.5"
githash = "09192c016185fa9af35d876f8364dc0af19a825a"


major = 0
minor = 5
bugfix = 0

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2019, 11, 25, 15, 34, 39)
debug = False

astropy_helpers_version = "3.2.2"
